"""Base tools to build simple buckets out of Quantitative and Qualitative features
for a binary classification model.
"""

from typing import Any, Union

from numpy import ndarray, sort
from pandas import isna


class GroupedList(list):
    """An ordered list that's extended with a dict."""

    def __init__(self, iterable: Union[ndarray, dict, list, tuple] = ()) -> None:
        """An ordered list that historizes its elements' content.

        Parameters
        ----------
        iterable : Union[ndarray, dict, list, tuple], optional
            List-like or GroupedList, by default ()
        """
        # TODO: move list to an attribute `order`?
        # case -1: iterable is an array
        if isinstance(iterable, ndarray):
            iterable = list(iterable)

        # case 0: iterable is the content dict
        if isinstance(iterable, dict):
            # storing ordered keys of the dict
            keys = list(iterable)

            # storing the values content per key
            self.content = dict(iterable.items())

            # checking that all values are only present once
            all_values = [val for _, values in iterable.items() for val in values]
            assert len(list(set(all_values))) == len(
                all_values
            ), "A value is present in several keys (groups)"

            # adding key to itself if it's not present in an other key
            keys_copy = keys[:]  # copying initial keys
            for key in keys_copy:
                # checking that the value is not comprised in an other key
                all_values = [
                    val
                    for iter_key, values in iterable.items()
                    for val in values
                    if key != iter_key
                ]
                if key not in all_values:
                    # checking that key is missing from its values
                    if key not in iterable[key]:
                        self.content.update({key: self.content[key] + [key]})
                # the key already is in another key (and its values are empty)
                # the key as already been grouped
                else:
                    self.content.pop(key)
                    keys.remove(key)

            # initiating the list with those keys
            super().__init__(keys)

        # case 1: copying a GroupedList
        elif hasattr(iterable, "content"):
            # initiating the list with the provided list of keys
            super().__init__(iterable)

            # copying values associated to keys
            self.content = dict(iterable.content.items())

        # case 2: initiating GroupedList from a list
        elif isinstance(iterable, list):
            # initiating the list with the provided list of keys
            super().__init__(iterable)

            # initiating the values with the provided list of keys
            self.content = {v: [v] for v in iterable}

    def get(self, key: Any) -> list[Any]:
        """List of values content in key

        Parameters
        ----------
        key : Any
            Group.

        Returns
        -------
        list[Any]
            Values content in key
        """

        # default to fing an element
        found = self.content.get(key)

        return found

    def group(self, discarded: Any, kept: Any) -> None:
        """Groups the discarded value with the kept value

        Parameters
        ----------
        discarded : Any
            Value to be grouped into the key `to_keep`.
        kept : Any
            Key value in which to group `discarded`.
        """

        # checking that those values are distinct
        if not is_equal(discarded, kept):
            # checking that those values exist in the list
            assert discarded in self, f"{discarded} not in list"
            assert kept in self, f"{kept} not in list"

            # accessing values content in each value
            content_discarded = self.content.get(discarded)
            content_kept = self.content.get(kept)

            # updating content dict
            self.content.update({kept: content_discarded + content_kept, discarded: []})

            # removing discarded from the list
            self.remove(discarded)

    def group_list(self, to_discard: list[Any], to_keep: Any) -> None:
        """Groups elements to_discard into values to_keep

        Parameters
        ----------
        to_discard : list[Any]
            Values to be grouped into the key `to_keep`.
        to_keep : Any
            Key value in which to group `to_discard` values.
        """

        for discarded, kept in zip(to_discard, [to_keep] * len(to_discard)):
            self.group(discarded, kept)

    def append(self, new_value: Any) -> None:
        """Appends a new_value to the GroupedList

        Parameters
        ----------
        new_value : Any
            New key to be added.
        """

        self += [new_value]
        self.content.update({new_value: [new_value]})

    def update(self, new_value: dict[Any, list[Any]]) -> None:  # TODO: not working as expected
        """Updates the GroupedList via a dict

        Parameters
        ----------
        new_value : dict[Any, list[Any]]
            Dict of key, values to updated `content` dict
        """

        # adding keys to the order if they are new values
        self += [key for key, _ in new_value.items() if key not in self]

        # updating content according to new_value
        self.content.update(new_value)

    def sort(self):
        """Sorts the values of the list and dict (if any, NaNs are last).

        Returns
        -------
        GroupedList
            Sorted GroupedList
        """
        # str values
        keys_str = [key for key in self if isinstance(key, str)]

        # non-str values
        keys_float = [key for key in self if not isinstance(key, str)]

        # sorting and merging keys
        keys = list(sort(keys_str)) + list(sort(keys_float))

        # recreating an ordered GroupedList
        sorted = GroupedList({k: self.get(k) for k in keys})

        return sorted

    def sort_by(self, ordering: list[Any]) -> None:
        """Sorts the values of the list and dict according to `ordering`, if any, NaNs are the last.

        Parameters
        ----------
        ordering : list[Any]
            Order used for ordering of the list of keys.

        Returns
        -------
        GroupedList
            Sorted GroupedList
        """

        # checking that all values are given an order
        assert all(
            o in self for o in ordering
        ), f"Unknown values in ordering: {str([v for v in ordering if v not in self])}"
        assert all(
            s in ordering for s in self
        ), f"Missing value from ordering: {str([v for v in self if v not in ordering])}"

        # ordering the content
        sorted = GroupedList({k: self.get(k) for k in ordering})

        return sorted

    def remove(self, value: Any) -> None:
        """Removes a value from the GroupedList

        Parameters
        ----------
        value : Any
            value to be removed
        """
        super().remove(value)
        self.content.pop(value)

    def pop(self, idx: int) -> None:
        """Pop a value from the GroupedList by index

        Parameters
        ----------
        idx : int
            Index of the value to be popped out
        """
        value = self[idx]
        self.remove(value)

    def get_group(self, value: Any) -> Any:
        """Returns the key (group) containing the specified value

        Parameters
        ----------
        value : Any
            Value for which to find the group.

        Returns
        -------
        Any
            Corresponding key (group)
        """

        found = [
            key
            for key, values in self.content.items()
            if any(is_equal(value, elt) for elt in values)
        ]

        if any(found):
            return found[0]

        return value

    def values(self) -> list[Any]:
        """All values content in all groups

        Returns
        -------
        list[Any]
            List of all values in the GroupedList
        """

        known = [value for values in self.content.values() for value in values]

        return known

    def contains(self, value: Any) -> bool:
        """Checks if a value is content in any group, also matches NaNs.

        Parameters
        ----------
        value : Any
            Value to search for

        Returns
        -------
        bool
            Whether the value is in the GroupedList
        """

        known_values = self.values()

        return any(is_equal(value, known) for known in known_values)

    def get_repr(self, char_limit: int = 6) -> list[str]:
        """Returns a representative list of strings of values of groups.

        Parameters
        ----------
        char_limit : int, optional
            Maximum number of character per string, by default 6

        Returns
        -------
        list[str]
            List of short str representation of the keys' values
        """

        # initiating list of group representation
        repr: list[str] = []

        # iterating over each group
        for group in self:
            # accessing group's values
            values = self.get(group)

            if len(values) == 0:  # case 0: there are no value in this group
                pass

            elif len(values) == 1:  # case 1: there is only one value in this group
                repr += [values[0]]

            elif len(values) == 2:  # case 2: two values in this group
                repr += [f"{values[1]}"[:char_limit] + " and " + f"{values[0]}"[:char_limit]]

            elif len(values) > 2:  # case 3: more than two values in this group
                repr += [f"{values[-1]}"[:char_limit] + " to " + f"{values[0]}"[:char_limit]]

        return repr


def is_equal(a: Any, b: Any) -> bool:
    """Checks if a and b are equal (NaN insensitive)

    Parameters
    ----------
    a : Any
        _description_
    b : Any
        _description_

    Returns
    -------
    bool
        _description_
    """

    # default equality
    equal = a == b

    # Case where a and b are NaNs
    if isna(a) and isna(b):
        equal = True

    return equal



"""Set of tests for quantitative_discretizers module."""

from pandas import DataFrame
from numpy import inf

from AutoCarver.discretizers.utils.quantitative_discretizers import QuantileDiscretizer


def test_quantile_discretizer(x_train: DataFrame):
    """Tests QuantileDiscretizer

    Parameters
    ----------
    x_train : DataFrame
        Simulated Train DataFrame
    """

    features = [
        "Quantitative",
        "Discrete_Quantitative",
        "Discrete_Quantitative_highnan",
        "Discrete_Quantitative_lownan",
        "Discrete_Quantitative_rarevalue",
    ]
    min_freq = 0.1

    discretizer = QuantileDiscretizer(features, min_freq, copy=True)
    x_discretized = discretizer.fit_transform(x_train)

    assert all(
        x_discretized.Quantitative.value_counts(normalize=True) == min_freq
    ), "Wrong quantiles"
    assert discretizer.values_orders["Discrete_Quantitative_highnan"] == [
        2.0,
        3.0,
        4.0,
        7.0,
        inf,
        "__NAN__",
    ], "NaNs should be added to the order"
    assert (
        "__NAN__" in x_discretized["Discrete_Quantitative_highnan"].unique()
    ), "NaNs should be filled with the str_nan value"
    assert discretizer.values_orders["Discrete_Quantitative_lownan"] == [
        1.0,
        2.0,
        3.0,
        4.0,
        6.0,
        inf,
        "__NAN__",
    ], "NaNs should not be grouped whatsoever"
    assert discretizer.values_orders["Discrete_Quantitative_rarevalue"] == [
        0.5,
        1.0,
        2.0,
        3.0,
        4.0,
        6.0,
        inf,
    ], "Wrongly associated rare values"



from codecs import open
from os import path

from setuptools import setup,find_packages

here = path.abspath(path.dirname(__file__))

with open(path.join(here, "README.md"), encoding="utf-8") as fh:
    long_description = "\\n" + fh.read()

setup(
    name="AutoCarver",
    version="5.0.5",
    author="Mario DEFRANCE",
    author_email="defrancemario@gmail.com",
    description="Automatic Bucketizing of Features with Optimal Association",
    long_description=long_description,
    long_description_content_type="text/markdown",
    url="https://github.com/mdefrance/AutoCarver",
    project_urls={
        "Bug Tracker": "https://github.com/mdefrance/AutoCarver/issues"
    },
    license="MIT",
    # install_requires= # TODO,
    packages=find_packages(),
    classifiers=[
        # ou 4 - Beta ou 5 - Production/Stable
        "Development Status :: 4 - Beta",
        "License :: OSI Approved :: MIT License",
        "Programming Language :: Python :: 3.9",
        "Programming Language :: Python :: 3.10",
        "Programming Language :: Python :: 3.11",
        "Operating System :: Unix",
        # "Operating System :: MacOS :: MacOS X",
        "Operating System :: Microsoft :: Windows",
    ],
    python_requires=">=3.7",
)



"""Sets of helper functions for GroupedListDiscretizer JSON serialization"""

from json import dumps, loads
from typing import Any, Union

from numpy import floating, inf, integer, isfinite

from .grouped_list import GroupedList


def convert_value_to_base_type(value: Any) -> Union[str, float, int]:
    """Converts a value to python's base types (str, int, float) for JSON serialization.

    Parameters
    ----------
    value : Any
        A value to serialize

    Returns
    -------
    Union[str, float, int]
        Serialized value
    """

    output = value  # str/float/int value
    if not isinstance(value, str) and not isfinite(value):  # numpy.inf value
        output = "numpy.inf"
    elif isinstance(value, integer):  # np.int value
        output = int(value)
    elif isinstance(value, floating):  # np.float value
        output = float(value)

    return output


def convert_value_to_numpy_type(value: Union[str, float, int]) -> Any:
    """Converts a list or a dict of lists values to numpy types for JSON deserialization.

    Parameters
    ----------
    value : Union[str, float, int]
        A value to deserialize

    Returns
    -------
    Any
        Deserialized value
    """
    output = value  # str/float/int value
    if value == "numpy.inf":  # numpy.inf value
        output = inf

    return output


def convert_values_to_base_types(
    iterable: Union[list[Any], dict[str, list[Any]]]
) -> Union[list[Union[str, int, float]], dict[str, list[Union[str, int, float]]]]:
    """Converts a list or a dict of lists values to python's base types (str, int, float) for JSON serialization.

    Parameters
    ----------
    iterable : Union[list[Any], dict[str, list[Any]]]
        List or dict of lists of values to serialize

    Returns
    -------
    Union[list[Union[str, int, float]], dict[str, list[Union[str, int, float]]]]
        List or dict of lists of values serialized
    """
    # list input
    output = None
    if isinstance(iterable, list):
        output = [convert_value_to_base_type(value) for value in iterable]
    # dict input
    elif isinstance(iterable, dict):
        output = {
            convert_value_to_base_type(key): convert_values_to_base_types(values)
            for key, values in iterable.items()
        }

    return output


def convert_values_to_numpy_types(
    iterable: Union[list[Union[str, int, float]], dict[str, list[Union[str, int, float]]]]
) -> Union[list[Any], dict[str, list[Any]]]:
    """Converts a list or a dict of lists values to numpy types for JSON deserialization.

    Parameters
    ----------
    iterable : Union[list[Union[str, int, float]], dict[str, list[Union[str, int, float]]]]
        List or dict of lists of values to deserialize

    Returns
    -------
    Union[list[Any], dict[str, list[Any]]]
        List or dict of lists of values deserialized
    """
    # list input
    output = None
    if isinstance(iterable, list):
        output = [convert_value_to_numpy_type(value) for value in iterable]
    # dict input
    elif isinstance(iterable, dict):
        output = {
            convert_value_to_numpy_type(key): convert_values_to_numpy_types(values)
            for key, values in iterable.items()
        }

    return output


def json_serialize_values_orders(values_orders: dict[str, GroupedList]) -> str:
    """JSON serializes a values_orders

    Parameters
    ----------
    values_orders : dict[str: GroupedList]
        values_orders to serialize

    Returns
    -------
    str
        JSON serialized values_orders
    """

    # converting values_orders to a json
    json_serialized_values_orders = {
        feature: {
            "order": convert_values_to_base_types(order),
            "content": convert_values_to_base_types(order.content),
        }
        for feature, order in values_orders.items()
    }

    return dumps(json_serialized_values_orders)


def json_deserialize_values_orders(json_serialized_values_orders: str) -> dict[str, GroupedList]:
    """JSON serializes a values_orders

    Parameters
    ----------
    json_serialized_values_orders : str
        JSON serialized values_orders

    Returns
    -------
    dict[str: GroupedList]
        values_orders deserialized
    """
    # converting to dict
    json_deserialized_values_orders = loads(json_serialized_values_orders)

    # converting to numpy type
    json_deserialized_values_orders = convert_values_to_numpy_types(json_deserialized_values_orders)

    # converting to GroupedList
    values_orders: dict[str, GroupedList] = {}
    for feature, content in json_deserialized_values_orders.items():
        # getting content from serialized dict
        feature_content: GroupedList = {}
        for value in content["order"]:
            content_key = value
            # float and int dict keys (converted in string by json.dumps)
            if not isinstance(value, str) and isfinite(value):
                content_key = str(value)
            feature_content.update({value: content["content"][content_key]})
        # saving up built GroupedList
        values_orders.update({feature: GroupedList(feature_content)})

    return values_orders
